#!/usr/bin/python

import os
import os.path
import string
from math import *
import sys
import wx
import dispatch
import qub.tree

# -------------------- enable one-line message dialogs -------------------------------

# open question: do you need to Destroy a component if you keep no references to it?

def wxMessage(*args):
	d= apply(wx.MessageDialog, args)
	r=d.ShowModal() # Shows it
	d.Destroy() # finally destroy it when finished.
	return r

# -------------------- cross-platform qub.tree clipboard support ---------------------

# wx clipboard is broken on win32
# we set it up here anyway, then prefer win32 implementation if available

# define CopyTree(node), CanPasteTree(), PasteTree()

qubtreeDataFormat = None

class qubtreeDataObject(wx.PyDataObjectSimple):
	def __init__(self, node=qub.tree.NullNode()):
		wx.PyDataObjectSimple.__init__(self, qubtreeDataFormat)
		self.node = node
	def GetDataSize(self):
		if self.node == None:
			return 0
		return len(self.node.getBytes())
	def GetDataHere(self):
		if self.node == None:
			return ""
		return self.node.getBytes()
	def SetData(self, str):
		try:
			node = qub.tree.ReadBytes(str)
			if not node.isNull:
				self.node = node
				return True
		except:
			pass
		return False
	def GetNode(self):
		return self.node

def InitClipboard():
	global qubtreeDataFormat
	if not qubtreeDataFormat:
		qubtreeDataFormat = wx.DataFormat(wx.DF_INVALID)
		qubtreeDataFormat.SetId("qubtree")

try:
	import win32clipboard
	import win32api
	treeClipFormat = win32clipboard.RegisterClipboardFormat('qubtree')
	nameClipFormat = win32clipboard.RegisterClipboardFormat('qubtreeName')
	def CopyTree(tree, hWnd=0):
		try:
			win32clipboard.OpenClipboard(hWnd)
			win32clipboard.EmptyClipboard()
			win32clipboard.SetClipboardData(treeClipFormat, tree.getBytes())
			win32clipboard.SetClipboardData(nameClipFormat, tree.name)
			win32clipboard.CloseClipboard()
			return True
		except win32api.error:
			return False
	def CanPasteTree(hWnd=0):
		try:
			win32clipboard.OpenClipboard(hWnd)
			result = bool(win32clipboard.IsClipboardFormatAvailable(treeClipFormat))
			win32clipboard.CloseClipboard()
			return result
		except win32api.error:
			return False
	def PasteTree(hWnd=0):
		result = qub.tree.NullNode()
		try:
			win32clipboard.OpenClipboard(hWnd)
			result = qub.tree.ReadBytes( win32clipboard.GetClipboardData(treeClipFormat) )
			win32clipboard.CloseClipboard()
		except win32api.error:
			pass
		return result
except:
	def CopyTree(tree, hWnd=0):
		if wx.TheClipboard.Open():
			wx.TheClipboard.SetData( qubtreeDataObject(self.GetPyData(self.GetSelection())) )
			wx.TheClipboard.Close()
			return True
		else:
			return False
	def CanPasteTree(hWnd=0):
		if wx.TheClipboard.Open():
			canPaste = wx.TheClipboard.IsSupported(qubtreeDataFormat)
			wx.TheClipboard.Close()
		else:
			canPaste = False
		return canPaste
	def PasteTree(hWnd=0):
		result = qub.tree.NullNode()
		if wx.TheClipboard.Open():
			data = qubtreeDataObject()
			wx.TheClipboard.GetData(data)
			result = data.GetNode()
			wx.TheClipboard.Close()
		return result

# ------------------- RowOfText < GridOfText < DataGrid -------------------------------

# wxPython doesn't include wx.Grid so this is what I came up with
# the idea was to use ready-made components to keep track of only the on-screen portion of data
# (which could otherwise be overwhelmingly large)
# the downside is that multi-select is not an option.
# potentially we could support clipboard as far as copy/paste all visible

GRID_ROW_HEIGHT = 16
GRID_COL_WIDTH_MIN = 50
GRID_LABEL_WIDTH_MAX = 50

class RowOfText(wx.Panel):
	def __init__(*args, **kw):
		apply(wx.Panel.__init__, args, kw)
		self = args[0]
		self.label = None
		self.texts = []
		self.LayoutChanged = lambda:None # dispatch.Dispatcher()
		wx.EVT_SIZE(self, self.OnSize)
	def OnSize(self,e):
		self.Layout()
	def SetLabel(self, s):
		if s:
			if not self.label:
				self.label = wx.StaticText(self, -1, style=wx.ALIGN_CENTER | wx.ST_NO_AUTORESIZE)
				self.Layout()
			self.label.SetLabel(s)
		else:
			if self.label:
				self.RemoveChild(self.label)
				self.label.Destroy()
				self.label = None
				self.Layout()
	def Layout(self):
		w, h = self.GetClientSize()
		lbl = 0
		if self.label:
			lbl = 1
		n = len(self.texts)
		if n>0 and w/(n+lbl) < GRID_COL_WIDTH_MIN:
			if self.label:
				self.label.SetPosition((0, 0))
				self.label.SetSize((GRID_COL_WIDTH_MIN, h))
			for i in range(n):
				j = i+lbl
				self.texts[i].SetPosition((j*GRID_COL_WIDTH_MIN, 0))
				self.texts[i].SetSize(((GRID_COL_WIDTH_MIN, h)))
		else:
			if self.label:
				self.label.SetPosition((0, 0))
				self.label.SetSize((GRID_LABEL_WIDTH_MAX, h))
			for i in range(n):
				self.texts[i].SetPosition((GRID_LABEL_WIDTH_MAX+(i*(w-lbl*GRID_LABEL_WIDTH_MAX))/n, 0))
				self.texts[i].SetSize((((i+1)*(w-lbl*GRID_LABEL_WIDTH_MAX))/n-(i*(w-lbl*GRID_LABEL_WIDTH_MAX))/n, h))
		self.LayoutChanged()
	def SetLength(self, n):
		diff = n - len(self.texts)
		if diff > 0:
			for i in range(diff):
				self.texts.append( wx.TextCtrl(self, -1) )
		elif diff < 0:
			for i in range(-diff):
				self.RemoveChild(self.texts[-i-1])
				self.texts[-i-1].Destroy()
			self.texts = self.texts[:diff]
		self.Layout()
	def GetLength(self):
		return len(self.texts)
	def GetValue(self, i):
		return self.texts[i].GetValue()
	def SetValue(self, i, v):
		self.texts[i].SetValue(v)

class GridOfText(wx.Panel):
	def __init__(*args, **kw):
		apply(wx.Panel.__init__, args, kw)
		self = args[0]
		self.rows = []
		self.labels = []
		self.MaxDimChanged = lambda maxnr,maxnc:None # dispatch.Dispatcher()
		wx.EVT_SIZE(self, self.OnSize)
		self.CalcMaxDim()
	def OnSize(self,e):
		self.Layout()
		maxnr, maxnc = self.GetMaxDim()
		self.CalcMaxDim()
		if maxnr != self.maxnr or maxnc != self.maxnc:
			self.MaxDimChanged(self.maxnr, self.maxnc)
	def CalcMaxDim(self):
		w, h = self.GetClientSize()
		self.maxnr = max(0, h/GRID_ROW_HEIGHT - 1)
		self.maxnc = max(0, w/GRID_COL_WIDTH_MIN - 1)
	def GetMaxDim(self):
		return (self.maxnr, self.maxnc)
	def Layout(self):
		w, h = self.GetClientSize()
		n = len(self.rows)
		for i in range(n):
			self.rows[i].SetPosition((0, (i+1)*GRID_ROW_HEIGHT))
			self.rows[i].SetSize((w, GRID_ROW_HEIGHT))
	def OnRowLayoutChanged(self):
		n = self.rows[0].GetLength()
		for i in range(n):
			self.labels[i].SetPosition(self.rows[0].texts[i].GetPosition())
			self.labels[i].SetSize(self.rows[0].texts[i].GetSize())
	def SetDim(self, nr, nc):
		diff = nr - len(self.rows)
		if diff > 0:
			for i in range(diff):
				self.rows.append( RowOfText(self, -1) )
				if len(self.rows) == 1:
					self.rows[0].LayoutChanged = self.OnRowLayoutChanged # was dispatcher +=
		elif diff < 0:
			for i in range(-diff):
				self.RemoveChild(self.rows[-i-1])
				self.rows[-i-1].Destroy()
			self.rows = self.rows[:diff]
		diff = nc - len(self.labels)
		if diff > 0:
			for i in range(diff):
				self.labels.append(wx.StaticText(self, -1, style=wx.ALIGN_CENTER | wx.ST_NO_AUTORESIZE))
		elif diff < 0:
			for i in range(-diff):
				self.RemoveChild(self.labels[-i-1])
				self.labels[-i-1].Destroy()
			self.labels = self.labels[:diff]
		for i in range(nr):
			self.rows[i].SetLength(nc)
		self.Layout()
	def GetDim(self):
		return (len(self.rows), len(self.rows) and self.rows[0].GetLength())
	def GetValue(self, i, j):
		return self.rows[i].GetValue(j)
	def SetValue(self, i, j, v):
		self.rows[i].SetValue(j, v)



class DataGrid(GridOfText):
	def __init__(self, parent, id):
		GridOfText.__init__(self, parent, id) #style: no resize
		self.node = qub.tree.NullNode()
		# clipboard? yeah good luck this grid doesn't multi-select
		self.MaxDimChanged = self.OnMaxDimChanged # was dispatcher +=
		wx.EVT_SCROLLWIN_THUMBRELEASE(self, self.OnScrollThumb)
		wx.EVT_SCROLLWIN_THUMBTRACK(self, self.OnScrollThumb)
		wx.EVT_SCROLLWIN_TOP(self, self.OnScrollTop)
		wx.EVT_SCROLLWIN_BOTTOM(self, self.OnScrollBottom)
		wx.EVT_SCROLLWIN_LINEUP(self, self.OnScrollLineUp)
		wx.EVT_SCROLLWIN_LINEDOWN(self, self.OnScrollLineDown)
		wx.EVT_SCROLLWIN_PAGEUP(self, self.OnScrollPageUp)
		wx.EVT_SCROLLWIN_PAGEDOWN(self, self.OnScrollPageDown)
	def OnMaxDimChanged(self, maxnr, maxnc):
		if not self.node.isNull:
			self.Commit()
			self.AdjustScroll()
			self.Display()
	def OnScrollThumb(self,e):
		i, j = self.i0, self.j0
		if e.GetOrientation() == wx.VERTICAL:
			i = e.GetPosition()
		else:
			j = e.GetPosition()
		self.ScrollTo(i, j)
	def OnScrollTop(self,e):
		if e.GetOrientation() == wx.VERTICAL:
			self.ScrollTo(0, self.j0)
		else:
			self.ScrollTo(self.i0, 0)
	def OnScrollBottom(self,e):
		if e.GetOrientation() == wx.VERTICAL:
			self.ScrollTo(self.node.data.rows-1, self.j0)
		else:
			self.ScrollTo(self.i0, self.node.data.cols-1)
	def OnScrollLineUp(self,e):
		if e.GetOrientation() == wx.VERTICAL:
			self.ScrollTo(max(0,self.i0-1), self.j0)
		else:
			self.ScrollTo(self.i0, max(0,self.j0-1))
	def OnScrollLineDown(self,e):
		if e.GetOrientation() == wx.VERTICAL:
			self.ScrollTo(min(self.node.data.rows-1,self.i0+1), self.j0)
		else:
			self.ScrollTo(self.i0, min(self.node.data.cols-1,self.j0+1))
	def OnScrollPageUp(self,e):
		nr, nc = self.GetDim()
		if e.GetOrientation() == wx.VERTICAL:
			self.ScrollTo(max(0,self.i0-(nr-1)), self.j0)
		else:
			self.ScrollTo(self.i0, max(0,self.j0-(nc-1)))
	def OnScrollPageDown(self,e):
		nr, nc = self.GetDim()
		if e.GetOrientation() == wx.VERTICAL:
			self.ScrollTo(min(self.node.data.rows-1,self.i0+(nr-1)), self.j0)
		else:
			self.ScrollTo(self.i0, min(self.node.data.cols-1,self.j0+(nc-1)))
	def ScrollTo(self, i, j):
		if self.i0 != i or self.j0 != j:
			self.Commit()
			self.i0 = i
			self.j0 = j
			self.AdjustScroll()
			self.Display()
	def SetNode(self, node):
		self.node = node
		self.i0, self.j0 = 0, 0
		self.AdjustScroll()
		self.Display()
	def AdjustScroll(self):
		data = self.node.data
		maxnr, maxnc = self.GetMaxDim()
		nr, nc = min(maxnr, data.rows), min(maxnc, data.cols)
		self.i0 = min(data.rows-nr, self.i0)
		self.j0 = min(data.cols-nc, self.j0)
		self.SetScrollbar(wx.VERTICAL, self.i0, nr, data.rows)
		self.SetScrollbar(wx.HORIZONTAL, self.j0, nc, data.cols)
		self.SetDim(nr, nc)
	def Display(self):
		node = self.node
		if node.data.count == 0:
			return
		nr, nc = self.GetDim()
		i0, j0 = self.i0, self.j0
		if not node.data.preload:
			node.data.loadRows(i0, i0+nr-1)
		for i in range(nr):
			row = node.data.row(i0+i)
			self.rows[i].SetLabel(str(i0+i+1))
			for j in range(nc):
				self.SetValue(i, j, str(row[j0+j]))
		for i in range(nc):
			self.labels[i].SetLabel(str(j0+i+1))
	def Commit(self):
		if self.node.data.type in (qub.tree.QTR_TYPE_FLOAT, qub.tree.QTR_TYPE_DOUBLE):
			format = float
		else:
			format = int
		j0 = self.j0
		nr, nc = self.GetDim()
		for i in range(nr):
			row = self.node.data.row(self.i0+i)
			for j in range(nc):
				try:
					row[j0+j] = format(self.GetValue(i, j))
				except Exception, e:
					pass # or complain?
		
# --------------------- NodeEditor dialog -------------------------
		
typestrs = ['None',
			'Unknown',
			'Pointer',
			'String',
			'8 bit unsigned',
			'8 bit int',
			'16 bit unsigned',
			'16 bit int',
			'32 bit unsigned',
			'32 bit int',
			'64 bit unsigned',
			'64 bit int',
			'32 bit float',
			'64 bit float',
			'80 bit float']
typeofstr = {}
for i,s in enumerate(typestrs):
	typeofstr[s] = i

typesnotdisplayed = {qub.tree.QTR_TYPE_EMPTY   : '',
					 qub.tree.QTR_TYPE_UNKNOWN : 'unknown data',
					 qub.tree.QTR_TYPE_POINTER : 'pointer data'}
typesunknown = [qub.tree.QTR_TYPE_UNKNOWN,
				qub.tree.QTR_TYPE_POINTER]
typesnumeric = [qub.tree.QTR_TYPE_UCHAR,
				qub.tree.QTR_TYPE_CHAR,
				qub.tree.QTR_TYPE_USHORT,
				qub.tree.QTR_TYPE_SHORT,
				qub.tree.QTR_TYPE_UINT,
				qub.tree.QTR_TYPE_INT,
				qub.tree.QTR_TYPE_ULONG,
				qub.tree.QTR_TYPE_LONG,
				qub.tree.QTR_TYPE_FLOAT,
				qub.tree.QTR_TYPE_DOUBLE,
				qub.tree.QTR_TYPE_LDOUBLE]
			
			
class NodeEditor(wx.Dialog):
	ID_NAME = 150
	ID_COMMENT = 151
	ID_TYPE = 152
	ID_ROWS = 153
	ID_COLS = 154
	ID_PRELOAD = 155
	ID_STRDATA = 156
	
	def __init__(self, parent, id, pos, size, node):
		wx.Dialog.__init__(self, parent, id, "QuB Tree Editor: %s"%node.name, pos, size)
		self.node = node

		self.sizer=wx.BoxSizer(wx.VERTICAL)
		self.sizer.AddSpacer((8,8))

		line = wx.BoxSizer(wx.HORIZONTAL)
		line.AddSpacer((8,8))
		line.Add( wx.StaticText(self, -1, " Name: "), 0, wx.EXPAND )
		self.txtName = wx.TextCtrl(self, self.ID_NAME, node.name)
		line.Add( self.txtName, 1, wx.EXPAND )
		line.Add( wx.StaticText(self, -1, " Comment: "), 0, wx.EXPAND )
		self.txtComment = wx.TextCtrl(self, self.ID_COMMENT, node.lineComment)
		line.Add( self.txtComment, 1, wx.EXPAND )
		line.AddSpacer((8,8))
		self.sizer.Add( line, 0, wx.EXPAND )
		self.sizer.AddSpacer((3,3))

		self.typeSizer = wx.BoxSizer(wx.HORIZONTAL)
		self.typeSizer.AddSpacer((8,8))
		self.typeSizer.Add( wx.StaticText(self, -1, " Type: "), 0, wx.EXPAND )
		self.mnuType = wx.ComboBox(self, self.ID_TYPE, style=wx.CB_DROPDOWN|wx.CB_READONLY,
								   value = typestrs[node.data.type], choices = typestrs)
		wx.EVT_COMBOBOX(self, self.ID_TYPE, self.OnSetType)
		self.typeSizer.Add( self.mnuType, 0, wx.EXPAND )
		self.dimSizer = wx.BoxSizer(wx.HORIZONTAL)
		self.dimSizer.Add( wx.StaticText(self, -1, " Rows: "), 0, wx.EXPAND )
		self.txtRows = wx.TextCtrl(self, self.ID_ROWS, str(node.data.rows), style=wx.TE_PROCESS_ENTER)
		wx.EVT_TEXT_ENTER(self, self.ID_ROWS, self.OnSetRows)
		wx.EVT_KILL_FOCUS(self.txtRows, self.OnSetRows)
		self.dimSizer.Add( self.txtRows, 0, wx.EXPAND )
		self.dimSizer.Add( wx.StaticText(self, -1, " Cols: "), 0, wx.EXPAND )
		self.txtCols = wx.TextCtrl(self, self.ID_COLS, str(node.data.cols), style=wx.TE_PROCESS_ENTER)
		wx.EVT_TEXT_ENTER(self, self.ID_COLS, self.OnSetCols)
		wx.EVT_KILL_FOCUS(self.txtCols, self.OnSetCols)
		self.dimSizer.Add( self.txtCols, 0, wx.EXPAND )
		self.chkPreload = wx.CheckBox(self, self.ID_PRELOAD, " Preload: ", style=wx.ALIGN_RIGHT)
		self.chkPreload.SetValue(node.data.preload)
		wx.EVT_CHECKBOX(self, self.ID_PRELOAD, self.OnCheckPreload)
		self.dimSizer.Add( self.chkPreload, 0, wx.EXPAND )
		self.typeSizer.Add( self.dimSizer, 1, wx.EXPAND )
		self.typeSizer.AddSpacer((8,8))
		self.sizer.Add( self.typeSizer, 0, wx.EXPAND )
		self.sizer.AddSpacer((3,3))

		self.dataSizer = wx.BoxSizer(wx.HORIZONTAL)
		self.dataSizer.AddSpacer((8,8))
		self.noData = wx.StaticText(self, -1, "", style=wx.ALIGN_CENTER|wx.ST_NO_AUTORESIZE)
		self.dataSizer.Add(self.noData, 1, wx.EXPAND)
		self.txtData = wx.TextCtrl(self, self.ID_STRDATA, style=wx.TE_MULTILINE|wx.TE_PROCESS_TAB)
		self.dataSizer.Add(self.txtData, 1, wx.EXPAND)
		self.grdData = DataGrid(self, -1)
		self.dataSizer.Add(self.grdData, 1, wx.EXPAND)
		self.dataSizer.AddSpacer((8,8))
		self.sizer.Add( self.dataSizer, 1, wx.EXPAND )
		self.sizer.AddSpacer((3,3))

		self.Display()
		
		# OK/Close button?
		wx.EVT_CLOSE(self, self.OnClose)

		self.SetSizerAndFit(self.sizer)
		self.SetSize(size)
		#modal or x close
		#commit, modify node and/or display settings, redisplay

	def Commit(self):
		node = self.node
		# bad to mess with name of root on disk:
		if (not node.path) or (not node.parent.isNull):
			node.name = str(self.txtName.GetValue())
		else:
			self.txtName.SetValue(node.name)
		node.lineComment = str(self.txtComment.GetValue())
		if node.data.type == qub.tree.QTR_TYPE_STRING:
			node.data = str(self.txtData.GetValue())
		elif not (node.data.type in typesnotdisplayed):
			self.grdData.Commit()
	def Display(self):
		node = self.node
		if node.data.type in typesnotdisplayed.keys():
			self.grdData.SetNode(qub.tree.NullNode())
			self.noData.SetLabel(typesnotdisplayed[node.data.type])
			self.dataSizer.Show(self.noData, True)
			self.dataSizer.Show(self.txtData, False)
			self.dataSizer.Show(self.grdData, False)
			self.typeSizer.Show(self.dimSizer, False)
		elif node.data.type == qub.tree.QTR_TYPE_STRING:
			self.grdData.SetNode(qub.tree.NullNode())
			self.txtData.SetValue(str(node.data))
			self.dataSizer.Show(self.noData, False)
			self.dataSizer.Show(self.txtData, True)
			self.dataSizer.Show(self.grdData, False)
			self.typeSizer.Show(self.dimSizer, False)
		else:
			self.grdData.SetNode(node)
			self.dataSizer.Show(self.noData, False)
			self.dataSizer.Show(self.txtData, False)
			self.dataSizer.Show(self.grdData, True)
			self.typeSizer.Show(self.dimSizer, True)
		self.Layout()
		self.txtRows.SetValue(str(self.node.data.rows))
		self.txtCols.SetValue(str(self.node.data.cols))

	def OnClose(self,e):
		self.Commit()
		self.EndModal(0)
	def OnSetType(self,e):
		self.Commit()
		oldtype = self.node.data.type
		newtype = typeofstr[self.mnuType.GetValue()]
		if newtype == oldtype:
			pass
		elif newtype in typesunknown or oldtype in typesunknown:
			self.mnuType.SetValue(typestrs[oldtype])
		elif newtype == qub.tree.QTR_TYPE_EMPTY:
			self.node.data.clear()
		elif newtype == qub.tree.QTR_TYPE_STRING:
			self.node.data = str(self.node.data) or " "
		else: # newtype in typesnumeric
			data = self.node.data
			copy = self.node.clone(False).data
			if oldtype == qub.tree.QTR_TYPE_STRING:
				data.setup(newtype, len(copy), 1)
				for i in range(len(copy)):
					data[i] = ord(copy[i])
			elif oldtype == qub.tree.QTR_TYPE_EMPTY:
				data.setup(newtype, 1, 1)
			else:
				data.setup(newtype, copy.rows, copy.cols)
				try:
					for i in range(len(copy)):
						data[i] = copy[i]
				except:
					pass
		self.Display()
	def OnSetRows(self,e):
		try:
			newnr = int(self.txtRows.GetValue())
			self.Commit()
			self.node.data.resize( newnr )
			self.Display()
		except:
			self.txtRows.SetValue(str(self.node.data.rows))
	def OnSetCols(self,e):
		try:
			newnc = int(self.txtCols.GetValue())
			if newnc == self.node.data.cols:
				return
			self.Commit()
			data = self.node.data
			oldnc = data.cols
			copy = self.node.clone(False).data
			data.setup(data.type, data.rows, newnc)
			data.loadRows()
			copy.loadRows()
			for i in range(data.rows):
				datarow = data.row(i)
				copyrow = copy.row(i)
				for j in range(min(newnc, oldnc)):
					datarow[j] = copyrow[j]
			self.Display()
		except:
			self.txtCols.SetValue(str(self.node.data.cols))
	def OnCheckPreload(self,e):
		self.node.data.preload = self.chkPreload.GetValue()


# -------------------------- TreeCtrl -----------------------------
		

def DefCaption(maxshown):
	def Caption(node):
		name, type, count = node.name, node.data.type, len(node.data)
		if node.lineComment:
			comment = " (%s)" % node.lineComment
		else:
			comment = ""
		if 0 == count:
			return "%s%s" % (name, comment)
		elif type == qub.tree.QTR_TYPE_STRING:
			return "%s: %s%s" % (name, node.data, comment)
		elif node.data.loaded[0] != 0 or type == qub.tree.QTR_TYPE_UNKNOWN or type == qub.tree.QTR_TYPE_POINTER:
			return "%s: <...>%s" % (name, comment)
		else:
			if type == qub.tree.QTR_TYPE_DOUBLE or type == qub.tree.QTR_TYPE_FLOAT or type == qub.tree.QTR_TYPE_LDOUBLE:
				format = lambda x: "%f" % x
			else:
				format = lambda x: "%d" % x
			shown = min(count, 10)
			if count > 10:
				more = ', ...'
			else:
				more = ''
			return "%s: %s%s%s" % (name, string.join([format(node.data[i]) for i in range(shown)], ", "), more, comment)
	return Caption

class TreeCtrl(wx.TreeCtrl):
	def __init__(self, parent, id, root, Caption=DefCaption(maxshown=10)):
		wx.TreeCtrl.__init__(self, parent, id)
		self.root = root
		self.Caption = Caption

		InitClipboard()		

		self.popup = wx.Menu()
		self.ID_EDIT = 101
		self.popup.Append(self.ID_EDIT, "&Edit"," Edit this node")
		wx.EVT_MENU(self, self.ID_EDIT, self.OnEdit) # attach the menu-event ID_ABOUT to the
		self.popup.AppendSeparator()
		self.ID_INS = 102
		self.popup.Append(self.ID_INS, "&Insert after"," Insert a new node after this node")
		wx.EVT_MENU(self, self.ID_INS, self.OnInsert) # attach the menu-event ID_ABOUT to the
		self.ID_INS_CHILD = 103
		self.popup.Append(self.ID_INS_CHILD, "Insert c&hild"," Insert a new child node")
		wx.EVT_MENU(self, self.ID_INS_CHILD, self.OnInsertChild) # attach the menu-event ID_ABOUT to the
		self.popup.AppendSeparator()
		self.ID_CUT = 104
		self.popup.Append(self.ID_CUT, "C&ut","")
		wx.EVT_MENU(self, self.ID_CUT, self.OnCut) # attach the menu-event ID_ABOUT to the
		self.ID_COPY = 105
		self.popup.Append(self.ID_COPY, "&Copy","")
		wx.EVT_MENU(self, self.ID_COPY, self.OnCopy) # attach the menu-event ID_ABOUT to the
		self.ID_PASTE = 106
		self.popup.Append(self.ID_PASTE, "&Paste after"," Paste after this node")
		wx.EVT_MENU(self, self.ID_PASTE, self.OnPaste) # attach the menu-event ID_ABOUT to the
		self.ID_PASTE_CHILD = 107
		self.popup.Append(self.ID_PASTE_CHILD, "Paste child"," Paste as a child of this node")
		wx.EVT_MENU(self, self.ID_PASTE_CHILD, self.OnPasteChild) # attach the menu-event ID_ABOUT to the
		self.ID_DEL = 108
		self.popup.Append(self.ID_DEL, "&Delete"," Delete this node")
		wx.EVT_MENU(self, self.ID_DEL, self.OnDelete) # attach the menu-event ID_ABOUT to the
		wx.EVT_CONTEXT_MENU(self, self.OnPopup)
		wx.EVT_TREE_ITEM_ACTIVATED(self, id, self.OnEdit)
		self.Init()
	def AddChildren(self, node, id):
		child = node.child
		while not child.isNull:
			if child.name != "QTR_LINE_COMMENT":
				id_child = self.AppendItem(id, self.Caption(child), data=wx.TreeItemData(child))
				self.AddChildren(child, id_child)
			child = child.sibling
	def Init(self):
		self.id_root = self.AddRoot(self.Caption(self.root), data=wx.TreeItemData(self.root))
		self.AddChildren(self.root, self.id_root)
		self.Expand(self.id_root)
	def Insert(self,sib):
		id = self.GetSelection()
		node = self.GetPyData(id)
		node.parent.insert(sib, node)
		id_parent = self.GetItemParent(id)
		id_sib= self.InsertItem(id_parent, id, self.Caption(sib), data=wx.TreeItemData(sib))
		self.AddChildren(sib, id_sib)
	def InsertChild(self,ch):
		id = self.GetSelection()
		node = self.GetPyData(id)
		node.insert(ch)
		id_ch = self.InsertItemBefore(id, 0, self.Caption(ch), data=wx.TreeItemData(ch))
		self.AddChildren(ch, id_ch)
		if not self.IsExpanded(id):
			self.Expand(id)
	def OnPopup(self,e):
		id = self.GetSelection()
		valid = id.IsOk()
		if valid:
			node = self.GetPyData(id)
			parented = not node.parent.isNull
			canPaste = bool(CanPasteTree())
		else:
			parented = False
			canPaste = False
		self.popup.Enable(self.ID_EDIT, valid)
		self.popup.Enable(self.ID_INS, parented)
		self.popup.Enable(self.ID_INS_CHILD, valid)
		self.popup.Enable(self.ID_CUT, parented)
		self.popup.Enable(self.ID_COPY, valid)
		self.popup.Enable(self.ID_PASTE, parented and canPaste)
		self.popup.Enable(self.ID_PASTE_CHILD, valid and canPaste)
		self.popup.Enable(self.ID_DEL, parented)

		self.PopupMenu(self.popup)
	def OnEdit(self,e):
		id = self.GetSelection()
		node = self.GetPyData(id)
		editor = NodeEditor(self, -1, self.GetParent().ClientToScreen(self.GetPosition()), self.GetSize(), node)
		editor.ShowModal()
		editor.Destroy()
		self.SetItemText(id, self.Caption(node))
	def OnInsert(self,e):
		self.Insert(qub.tree.Node('untitled'))
	def OnInsertChild(self,e):
		self.InsertChild(qub.tree.Node('untitled'))
	def OnCut(self,e):
		if CopyTree( self.GetPyData(self.GetSelection()) ):
			self.OnDelete(e)
	def OnCopy(self,e):
		CopyTree( self.GetPyData(self.GetSelection()) )
	def OnPaste(self,e):
		pasted = PasteTree()
		if not pasted.isNull:
			self.Insert(pasted)
	def OnPasteChild(self,e):
		pasted = PasteTree()
		if not pasted.isNull:
			self.InsertChild(pasted)
	def OnDelete(self,e):
		id = self.GetSelection()
		node = self.GetPyData(id)
		self.Delete(id)
		node.parent.remove(node)


# -------------------------- MainWindow -----------------------------------

ID_NEW=102
ID_OPEN=103
ID_SAVE=104
ID_SAVEAS=105

class MainWindow(wx.Frame):
	def __init__(self,parent,id):
		wx.Frame.__init__(self,parent,wx.ID_ANY, 'QuB Tree Editor: <untitled>', size = (500, 350),

										style=wx.DEFAULT_FRAME_STYLE|
										wx.NO_FULL_REPAINT_ON_RESIZE)
		self.CreateStatusBar() # A StatusBar in the bottom of the window
		self.dirname = ''
		self.filename = ''
		wx.EVT_CLOSE(self, self.OnClose)
		# Setting up the menu.
		filemenu= wx.Menu()
		self.filemenu = filemenu
		filemenu.Append(wx.ID_ABOUT, "&About"," Information about this program")
		wx.EVT_MENU(self, wx.ID_ABOUT, self.OnAbout) # attach the menu-event ID_ABOUT to the
		filemenu.AppendSeparator()
		filemenu.Append(ID_NEW, "&New"," Create an empty QuB tree")
		wx.EVT_MENU(self, ID_NEW, self.OnNew) # attach the menu-event ID_ABOUT to the
		filemenu.Append(ID_OPEN, "&Open"," Open a QuB tree from disk")
		wx.EVT_MENU(self, ID_OPEN, self.OnOpen) # attach the menu-event ID_ABOUT to the
		filemenu.Append(ID_SAVE, "&Save"," Save to disk")
		wx.EVT_MENU(self, ID_SAVE, self.OnSave) # attach the menu-event ID_ABOUT to the
		filemenu.Append(ID_SAVEAS, "&Save As"," Save to a new file")
		wx.EVT_MENU(self, ID_SAVEAS, self.OnSaveAs) # attach the menu-event ID_ABOUT to the
		filemenu.AppendSeparator()
		filemenu.Append(wx.ID_EXIT,"E&xit"," Terminate the program")
		wx.EVT_MENU(self, wx.ID_EXIT, self.OnExit)   # attach the menu-event ID_EXIT to the
		# Creating the menubar.
		menuBar = wx.MenuBar()
		menuBar.Append(filemenu,"&File") # Adding the "filemenu" to the MenuBar
		self.SetMenuBar(menuBar)  # Adding the MenuBar to the Frame content.
		wx.EVT_MENU_OPEN(self, self.OnMenuOpen)

		self.editor = TreeCtrl(self, 1, qub.tree.Node('Untitled'))
		self.sizer=wx.BoxSizer(wx.VERTICAL)
		self.sizer.Add(self.editor,1,wx.EXPAND)
		#Layout sizers
		self.SetSizerAndFit(self.sizer)
		self.SetSize((600, 400))

		self.Show(True)
	def NewEditor(self, tree):
		self.sizer.Detach(self.editor)
		self.editor.Destroy()
		self.editor = TreeCtrl(self, 1, tree)
		self.sizer.Add(self.editor,1,wx.EXPAND)
		self.sizer.Layout()
	def ConfirmDiscard(self):
		if (self.editor.root.modified
		and ((not self.editor.root.child.isNull)  # i.e. if it's not the original blank node
		   or len(self.editor.root.data))):
			response = wxMessage(self, "Save Changes?", "Save changes to %s?" % self.filename, wx.YES | wx.NO | wx.CANCEL)
			if response == wx.ID_YES:
				if self.editor.root.path:
					if not self.editor.root.save():
						wxMessage(self, "Error", "Can't save %s" % self.filename, wx.OK)
						response = wx.ID_CANCEL
				else:
					if not self.SaveAs():
						response = wx.ID_CANCEL
			return response != wx.ID_CANCEL
		return True
	def OnClose(self,e):
		if e.CanVeto():
			if self.ConfirmDiscard():
				self.Destroy()
			else:
				e.Veto()
		else:
			self.Destroy()
	def OnMenuOpen(self,e):
		self.filemenu.Enable(ID_SAVE, bool(self.editor.root.path) and self.editor.root.modified)
	def OnAbout(self,e):
		wxMessage( self, "QuB tree editor\nwww.qub.buffalo.edu/qubdoc","About QuB Tree Editor", wx.OK)
	def OnNew(self,e):
		self.New()
	def OnOpen(self,e):
		""" Open a file"""
		if self.ConfirmDiscard():
			dlg = wx.FileDialog(self, "Choose a file", self.dirname, "", "*.*", wx.OPEN)
			result = dlg.ShowModal()
			if result == wx.ID_OK:
				self.filename=dlg.GetFilename()
				self.dirname=dlg.GetDirectory()
				self.Open(os.path.join(self.dirname, self.filename))
			dlg.Destroy()
	def OnSave(self,e):
		self.Save()
	def OnSaveAs(self,e):
		self.SaveAs()
	def New(self):
		if self.ConfirmDiscard():
			self.NewEditor(qub.tree.Node('Untitled'))
			self.SetTitle("QuB Tree Editor: <untitled>")
			self.filename = ''
	def Open(self, path):
		tree = qub.tree.Open(path)
		if not tree.isNull:
			self.NewEditor(tree)
			self.SetTitle("QuB Tree Editor: %s" % path)
	def Save(self):
		if not self.editor.root.save():
			wxMessage(self, "Error", "Can't save %s" % self.filename, wx.OK)
	def SaveAs(self):
		dlg = wx.FileDialog(self, "Save as", self.dirname, self.filename, "*.*", wx.SAVE | wx.OVERWRITE_PROMPT)
		result = dlg.ShowModal() == wx.ID_OK
		if result:
			self.filename = dlg.GetFilename()
			self.dirname = dlg.GetDirectory()
			result = self.editor.root.saveAs(os.path.join(self.dirname, self.filename))
			if not result:
				wxMessage(self, "Error", "Can't save %s" % self.filename, wx.OK)
			else:
				self.SetTitle("QuB Tree Editor: %s" % os.path.join(self.dirname, self.filename))
		dlg.Destroy()
		return result
	def OnExit(self,e):
		self.Close(True)  # Close the frame.

if __name__ == '__main__':
	app = wx.PySimpleApp()
	frame=MainWindow(None,-1)
	if len(sys.argv) > 1:
		frame.Open(sys.argv[1])
	app.MainLoop()


